@@ -3,11 +3,6 @@
  * See the file COPYING for copying conditions.
  */
 
-/*
- * TODO: Move most of the includes from kloak.c to here; right now this is a
- * very non-portable header file.
- */
-
 /*********************************/
 /* static defines, do not change */
 /*********************************/
@@ -15,11 +10,20 @@
 #define ASCII_UPPERCASE_START 65
 #define ASCII_LOWERCASE_START 97
 #define UNIX_SOCK_PATH_MAX 107
+/*
+ * See the scroll handling comment in
+ * queue_libinput_event_and_relocate_virtual_cursor() for details on why the
+ * scroll-related values below were chosen.
+ */
+#define SCROLL_UNITS_PER_TICK 120
+#define SCROLL_UNITS_PER_TICK_D 120.0
+#define SCROLL_ANGLE_PER_TICK 15
+#define SCROLL_ANGLE_TO_UNITS_FACTOR_D 8.0
 
 /*******************/
 /* tunable defines */
 /*******************/
-#define MAX_DRAWABLE_LAYERS 128
+#define MAX_SCREEN_COUNT 128
 #define CURSOR_RADIUS 15
 #define POLL_TIMEOUT_MS 1
 #define DEFAULT_MAX_DELAY_MS 100
@@ -41,8 +45,8 @@
 /*******************/
 
 /*
- * Defines a libinput device on the system by its device ointer and file name.
- * Intended for use in a doubly-linked list.
+ * Defines a libinput device on the system by its device pointer and file
+ * name. Intended for use in a doubly-linked list.
  */
 struct li_device_info {
   struct libinput_device *device;
@@ -116,27 +120,58 @@ struct coord {
 };
 
 /*
- * Defines a buffered input event. Two types of events are supported, mouse
- * movement events and libinput events. libinput events can be any arbitrary
- * event supported by libinput. Mouse movement events are defined as a cursor
- * position in compositor global space. Both kinds of events have a scheduled
- * release time. An `entries` field is included to allow buffered events to be
- * stored in a tail queue.
+ * Defines the type of a buffered input event.
+ */
+enum input_packet_type {
+  KLOAK_PACKET_TYPE_MOUSEMOVE,
+  KLOAK_PACKET_TYPE_MOUSESCROLL,
+  KLOAK_PACKET_TYPE_LIBINPUT
+};
+
+/*
+ * A container for storing the actual data in an input_packet, as explained
+ * below.
+ */
+union input_packet_data {
+  struct {
+    int32_t cursor_x;
+    int32_t cursor_y;
+  } mousemove;
+
+  struct {
+    int32_t vert_scroll_ticks;
+    int32_t horiz_scroll_ticks;
+  } mousescroll;
+
+  struct {
+    struct libinput_event *li_event;
+  } libinput;
+};
+
+/*
+ * Defines a buffered input event. Three types of events are supported, mouse
+ * movement events, scroll events, and libinput events.
+ *
+ * - Mouse movement events are defined as a cursor position in compositor
+ *   global space.
+ * - Scroll events are defined as a pair of scroll "tick" values, one for each
+ *   dimension (vertical and horizontal).
+ * - libinput events are raw events directly from libinput, intended for
+ *   release unmodified at a later time.
+ *
+ * All kinds of events have a scheduled release time. An `entries` field is
+ * included to allow buffered events to be stored in a tail queue.
+ *
+ * The actual data of the input packet is in the `data` field. The fields of
+ * `data` (mousemove, mousescroll, libinput) correspond to the
+ * `KLOAK_PACKET_TYPE_MOUSEMOVE`, `KLOAK_PACKET_TYPE_MOUSESCROLL`, and
+ * `KLOAK_PACKET_TYPE_LIBINPUT` packet types, respectively.
  */
 struct input_packet {
-  bool is_libinput;
-
-  /* libinput-specific bits */
-  struct libinput_event * li_event;
-  enum libinput_event_type li_event_type;
-
-  /* mouse movement bits */
-  int32_t cursor_x;
-  int32_t cursor_y;
-
-  /* generic bits */
+  enum input_packet_type packet_type;
   int64_t sched_time;
   TAILQ_ENTRY(input_packet) entries;
+  union input_packet_data data;
 };
 
 /*
@@ -153,12 +188,12 @@ struct disp_state {
   uint32_t seat_caps;
   bool seat_set;
   struct wl_keyboard *kb;
-  struct wl_output *outputs[MAX_DRAWABLE_LAYERS];
-  uint32_t output_names[MAX_DRAWABLE_LAYERS];
+  struct wl_output *outputs[MAX_SCREEN_COUNT];
+  uint32_t output_names[MAX_SCREEN_COUNT];
   struct zxdg_output_manager_v1 *xdg_output_manager;
-  struct zxdg_output_v1 *xdg_outputs[MAX_DRAWABLE_LAYERS];
-  struct output_geometry *output_geometries[MAX_DRAWABLE_LAYERS];
-  struct output_geometry *pending_output_geometries[MAX_DRAWABLE_LAYERS];
+  struct zxdg_output_v1 *xdg_outputs[MAX_SCREEN_COUNT];
+  struct output_geometry *output_geometries[MAX_SCREEN_COUNT];
+  struct output_geometry *pending_output_geometries[MAX_SCREEN_COUNT];
   int32_t global_space_width;
   int32_t global_space_height;
   int32_t pointer_space_x;
@@ -174,7 +209,7 @@ struct disp_state {
   struct xkb_state *xkb_state;
   char *old_kb_map_shm;
   int32_t old_kb_map_shm_size;
-  struct drawable_layer *layers[MAX_DRAWABLE_LAYERS];
+  struct drawable_layer *layers[MAX_SCREEN_COUNT];
 };
 
 /***************/
@@ -331,7 +366,7 @@ static uint32_t parse_uint32_arg(const char *arg_name, const char *val,
 static int32_t sleep_ms(int64_t ms);
 
 /*
- * Allocates a formatted string and returns a ointer to it. It is the caller's
+ * Allocates a formatted string and returns a pointer to it. It is the caller's
  * responsibility to free this when it is no longer in use.
  */
 static char *sgenprintf(const char *str, ...);
@@ -351,6 +386,13 @@ static void attach_input_device(const char *dev_name);
  */
 static void detach_input_device(const char *dev_name);
 
+/*
+ * Extracts a whole number of scroll ticks from a scroll accumulator,
+ * modifying the accumulator's value as appropriate to remove that number of
+ * ticks from it.
+ */
+static int32_t get_ticks_from_scroll_accum(double *accum_ptr);
+
 /********************/
 /* wayland handling */
 /********************/
@@ -453,28 +495,35 @@ static void damage_surface_enh(struct wl_surface *surface, int32_t x,
  * isn't one already queued, and update a queued mouse movement event to
  * reflect the current virtual cursor position otherwise.
  */
-static struct input_packet * update_virtual_cursor(void);
+static struct input_packet *update_virtual_cursor(void);
+
+/*
+ * Updates the scroll accumulators. If enough scroll movement has accumulated
+ * to scroll one or more ticks, this will create a new mouse scroll event if
+ * there isn't an appropriate one already queued, and update a queued scroll
+ * event to reflect the new scroll quantity otherwise.
+ */
+static struct input_packet *update_mouse_scroll(void);
 
 /*
  * Processes a libinput event, sending emulated input to the compositor as
  * appropriate.
  */
-static void handle_libinput_event(enum libinput_event_type ev_type,
-  struct libinput_event *li_event, uint32_t ts_milliseconds);
+static void handle_libinput_event(struct libinput_event *li_event,
+  uint32_t ts_milliseconds);
 
 /*
  * Tracks actively held down keys in the escape key list and terminates kloak
  * if all escape keys are actively being held down.
  */
-static void register_esc_combo_event(enum libinput_event_type li_event_type,
-  struct libinput_event *li_event);
+static void register_esc_combo_event(struct libinput_event *li_event);
 
 /*
  * Schedules a libinput event for future release to the compositor. As a side
  * effect, also redraws the virtual cursor if needed.
  */
 static void queue_libinput_event_and_relocate_virtual_cursor(
-  enum libinput_event_type li_event_type, struct libinput_event *li_event);
+  struct libinput_event *li_event);
 
 /*
  * Finds all queued input events that are ready to be released, and process
